"""Model a financial asset in the context of this library."""
from abc import ABC
from functools import cached_property
from typing import Literal

from requests import HTTPError

from vistafetch.constants import ONVISTA_API_BASE_URL
from vistafetch.model.asset import PriceData
from vistafetch.model.asset.financial_asset_type import FinancialAssetType
from vistafetch.model.base import VistaEntity
from vistafetch.session import api_session

__all__ = [
    "FinancialAsset",
]


class FinancialAsset(VistaEntity, ABC):
    """General description of a financial asset in the context of this library.

    This class models the shared attributes of a financial asset as
    returned from the API.cThese attributes are not comprehensive,
    it's currently only a selection of the most important ones.
    Thus, they may get extended in the future.

    Please note, that this class is only intended to serve as
    an abstract base class.The concrete financial assets as
    returned by the API still need to be defined. Thereby,
    it is required to overwrite the attribute `_type` accordingly.

    Attributes
    ----------
    display_type: display name of the asset type
    entity_type: name of the entity type as determined by the API,
        use values of `FinancialAssetType`
    isin: ISIN code of the asset
        (An International Securities Identification Number (ISIN)
        is a code that uniquely identifies a security globally
        for the purposes of facilitating clearing, reporting,
        and settlement of trades.)
    name: full name of the financial asset
    tiny_name: short name of the financial asset, preferably useful as display name
    wkn: WKN code of the financial asset (The 'Wertpapierkennnummer' (WKN) is a
        German securities identification code.)

    """

    _type: FinancialAssetType = FinancialAssetType.UNKNOWN

    display_type: str
    entity_type: Literal[_type.value]  # type: ignore
    isin: str
    name: str
    tiny_name: str
    wkn: str

    def __query_price_data(self) -> PriceData:
        if self._type == FinancialAssetType.UNKNOWN:
            raise NotImplementedError(
                "`price_data` is called directly on the "
                "abstract class `Financial Asset`. "
                "Please use a valid financial asset class."
            )
        response = api_session.get(
            f"{ONVISTA_API_BASE_URL}{self._type.value.lower()}s/ISIN:{self.isin}/snapshot"
        )
        try:
            response.raise_for_status()
        except HTTPError as e:
            raise RuntimeError(f"API does not return a valid response: {e}")

        response_dict = response.json()
        if "quote" not in response_dict.keys():
            raise ValueError(
                f"API response does not contain expected data: {response_dict}"
            )

        return PriceData.model_validate(response_dict["quote"])

    @cached_property
    def price_data(self) -> PriceData:
        """Get the price data available for this financial asset."""
        return self.__query_price_data()



"""Model a stock."""
from typing import Literal

from vistafetch.model.asset.financial_asset import FinancialAsset
from vistafetch.model.asset.financial_asset_type import FinancialAssetType

__all__ = [
    "Stock",
]


class Stock(FinancialAsset):
    """Models a stock within the scope of this library.

    A stock is a type of investment that represents ownership in a company.
    When one buys a stock, one is purchasing a small piece of
        that company, called a share.

    """

    _type = FinancialAssetType.STOCK

    entity_type: Literal[_type.value]  # type: ignore



"""Model financial information in the context of this library."""
from datetime import datetime

from pydantic import Field, field_validator

from vistafetch.constants import ONVISTA_API_TIMEZONE
from vistafetch.model.base import VistaEntity

__all__ = [
    "PriceData",
]


class PriceData(VistaEntity):
    """Price data of a financial asset.

    This class models price-related data of a financial asset.
    Alongside the currency, it provides different price parameters and
    the corresponding timestamp of measurement.

    Attributes
    ----------
        currency_symbol: symbol of the corresponding currency,
            in ISO notation as returned by the API, e.g., 'EUR'.
        datetime_high: timestamp of measurement for price value `high`, timezone-aware
        datetime_last: timestamp of measurement for price value `last`, timezone-aware
        datetime_low: timestamp of measurement for price value `low`, timezone-aware
        datetime_open: timestamp of measurement for price value `open`,
            timezone-aware
        high: The highest price at which the stock was traded during
            the day's trading session.
        low: The lowest price at which the stock was traded during
            the day's trading session.
        last: The last price at which the stock was traded before
            the market closed.
            This is also known as the closing price.
        open: The price at which the stock began trading at
            the beginning of the day's trading session.

    """

    currency_symbol: str = Field(alias="isoCurrency")
    datetime_high: datetime
    datetime_last: datetime
    datetime_low: datetime
    datetime_open: datetime
    high: float
    last: float
    low: float
    open: float

    @field_validator("datetime_high", "datetime_last", "datetime_low", "datetime_open")
    @classmethod
    def __ensure_timezone_awareness(cls, dt: datetime) -> datetime:
        if not isinstance(dt, datetime):
            raise ValueError(f"{dt} is not a datetime instance.")
        return dt.replace(tzinfo=ONVISTA_API_TIMEZONE)



"""Basic model definitions.

These are not defined to are used directly but to server for further modelling.

"""
from typing import Any, Dict

from pydantic import BaseModel, ConfigDict

__all__ = [
    "VistaEntity",
]


def _snake_to_camel_case(snake_case_string: str) -> str:
    """Convert a string in snake_case format to camelCase style.

    Args:
    ----
    snake_case_string: str
        string in snake_case format

    Returns:
    -------
    str

    """
    tokens = snake_case_string.split("_")

    return tokens[0] + "".join(t.title() for t in tokens[1:])


class VistaEntity(BaseModel):
    """Basic vista entity representation.

    This class serves as a common base class for several model classes of this library.

    Attributes
    ----------
        model_config: ConfigDict
            Config object to specify the behavior of the model class.
            Read more at [Pydantic's docs](https://docs.pydantic.dev/latest/usage/model_config/).

    """

    model_config = ConfigDict(alias_generator=_snake_to_camel_case, extra="allow")

    def as_json(self) -> str:
        """Return the entity as JSON string.

        This method returns the attributes of the entity data model as JSON string.
        It excludes additional parameters that are returned by the API but not modelled.
        Use `extra()` to get access to additional price-related data.

        Returns
        -------
            model_json: The model instance represented as JSON string
        """
        # By default, Pydantic includes additional parameters as well in the JSON dump
        # https://github.com/pydantic/pydantic/issues/6150
        # Therefore, we need to exclude them by hand
        return self.model_dump_json(
            exclude=set(self.model_extra.keys() if self.model_extra else [])
        )

    @property
    def extra(self) -> Dict[str, Any]:
        """Additional data returned by the API but not covered by this model.

        Extra data is exactly as returned by the API,
        no modifications are applied.

        Returns
        -------
            extra: dictionary containing additional data.
        """
        return self.model_extra if self.model_extra else {}



"""Search result as returned by the API when querying a financial asset."""
from datetime import datetime
from typing import List, Union

from pydantic import Field
from rich.console import Console
from rich.table import Table

from vistafetch.model.asset import Bond, Fund, Index, PreciousMetal, Stock
from vistafetch.model.asset.financial_asset import FinancialAsset
from vistafetch.model.base import VistaEntity

__all__ = [
    "SearchResult",
]


class SearchResult(VistaEntity):
    """Search result as returned by the API.

    This represents the return when searching for a financial asset via the API.

    Attributes
    ----------
        expires: timestamp until the search result is considered to be valid by the API.
        assets: list of financial assets that match the search term
        search_value: the actual value that has been searched for

    """

    expires: datetime
    assets: List[Union[Bond, Fund, Index, PreciousMetal, Stock]] = Field(
        alias="list", discriminator="entity_type"
    )
    search_value: str

    def get(self, index: int = 0) -> FinancialAsset:
        """Get a dedicated financial asset from the result.

        By default, the asset with index '0' is returned.

        Args:
        ----
        index: index of the asset to be returned, default: 0

        Returns:
        -------
        FinancialAsset

        Raises:
        ------
        AttributeError: in case the given index is not available
        RuntimeError: in case the search didn't return any match

        """
        if len(self.assets) == 0:
            raise RuntimeError(
                f"No results found for search term '{self.search_value}'"
            )
        try:
            return self.assets[index]
        except IndexError as e:
            raise AttributeError(
                f"The given index '{index}' is not available. "
                f"Maximum available index: '{len(self.assets)}'"
            ) from e

    def visualize(self) -> None:
        """Visualize the search result in the console.

        Returns
        -------
            None
        """
        console = Console()

        table = Table(title="Financial assets discovered")

        table.add_column("Index", justify="center")
        table.add_column("Name", justify="center")
        table.add_column("Asset Type", justify="center")
        table.add_column("ISIN", justify="center")

        for idx, asset in enumerate(self.assets):
            table.add_row(str(idx), asset.tiny_name, asset.entity_type, asset.isin)

        console.print(table)



"""Model an index."""
from typing import Literal

from vistafetch.model.asset.financial_asset import FinancialAsset
from vistafetch.model.asset.financial_asset_type import FinancialAssetType

__all__ = [
    "Index",
]


class Index(FinancialAsset):
    """Models an investment index within the scope of this library.

    A stock index is a measurement of the stock market that tracks
    the performance of a group of stocks or other assets.

    """

    _type = FinancialAssetType.INDEX

    entity_type: Literal[_type.value]  # type: ignore



"""Model a precious metal."""
from typing import Literal

from vistafetch.model.asset.financial_asset import FinancialAsset
from vistafetch.model.asset.financial_asset_type import FinancialAssetType

__all__ = [
    "PreciousMetal",
]


class PreciousMetal(FinancialAsset):
    """Models an investment index within the scope of this library.

    A precious metal is a rare metallic element that has high economic value
    due to various factors, including their scarcity, use in industrial processes,
    hedge against currency inflation, and role throughout history as a store of value.

    """

    _type = FinancialAssetType.METAL

    entity_type: Literal[_type.value]  # type: ignore



"""Client to fetch Onvista API."""
import logging
from typing import Dict, Optional

from vistafetch.logs import set_up_logging
from requests import HTTPError

from vistafetch.constants import ONVISTA_API_BASE_URL
from vistafetch.session import api_session
from vistafetch.model import SearchResult

__all__ = [
    "VistaFetchClient",
]

logger = logging.getLogger(name=__name__)


class VistaFetchClient:
    """Client to fetch financial data from the Onvista API.

    Args:
    ----
        client_headers: additional headers to be sent with every request, optional
    """

    def __init__(self, client_headers: Optional[Dict[str, str]] = None):
        set_up_logging()

        http_headers = {"Application": "application/json; charset=utf-8"}

        if client_headers:
            http_headers.update(client_headers)

        # set up a requests session
        # this allows to centrally determine the behavior of all requests made
        api_session.headers.update(http_headers)

        logger.debug(
            "Requests session has been configured with the following headers: \n"
            f"{http_headers}"
        )
        logger.info("Client has been initialized successfully.")

    @staticmethod
    def search_asset(
        search_term: str,
        max_candidates: Optional[int] = 5,
    ) -> SearchResult:
        """Search for a financial asset.

        Allow to search for specific financial assets.
        Search terms can be (parts of) the asset name, the ISIN, WKN, etc.

        Args:
        ----
        search_term: str
            the term to searched for
        max_candidates: Optional[int]
            maximum amount of returned assets

        Returns:
        -------
            SearchResult

        """
        response = api_session.get(
            f"{ONVISTA_API_BASE_URL}instruments/query?limit={max_candidates}&searchValue={search_term}"
        )
        try:
            response.raise_for_status()
        except HTTPError as e:
            raise RuntimeError(f"API does not return a valid response: {e}")
        logger.debug(f"Response retrieved from the API: {response.json()}")

        return SearchResult.model_validate(response.json())



"""Model a bond."""
from typing import Literal

from vistafetch.model.asset.financial_asset import FinancialAsset
from vistafetch.model.asset.financial_asset_type import FinancialAssetType

__all__ = [
    "Bond",
]


class Bond(FinancialAsset):
    """Models a bond within the scope of this library.

    A bond is a type of debt security that represents
    a loan made by an investor to a borrower,
    such as a government or corporation.
    When an investor buys a bond, they are essentially lending money to
    the issuer for a specified period of time,
    during which the issuer promises to pay the investor a fixed rate of interest.

    """

    _type = FinancialAssetType.BOND

    entity_type: Literal[_type.value]  # type: ignore



"""Model several financial assets."""

from vistafetch.model.asset.financial_data import PriceData
from vistafetch.model.asset.bond import Bond
from vistafetch.model.asset.fund import Fund
from vistafetch.model.asset.index import Index
from vistafetch.model.asset.metal import PreciousMetal
from vistafetch.model.asset.stock import Stock

__all__ = [
    "Bond",
    "Fund",
    "Index",
    "PreciousMetal",
    "PriceData",
    "Stock",
]



"""Model a fund."""
from typing import Literal

from vistafetch.model.asset.financial_asset import FinancialAsset
from vistafetch.model.asset.financial_asset_type import FinancialAssetType

__all__ = [
    "Fund",
]


class Fund(FinancialAsset):
    """Models an investment fund within the scope of this library.

    An investment fund is a financial product that pools money
    from multiple investors to buy a diversified portfolio of
    assets, such as stocks, bonds, real estate, and commodities.
    This comprises various types of investment funds,
    including mutual funds, exchange-traded funds (ETFs),
    and hedge funds, each with its own investment strategy and risk profile.

    """

    _type = FinancialAssetType.FUND

    entity_type: Literal[_type.value]  # type: ignore



"""Requests session to be commonly shared.

Attributes
----------
api_session: Session to communicate with the API

"""

from requests import Session

__all__ = [
    "api_session",
]


api_session = Session()



"""Collection of model classes.

All model classes are defined using Pydantic.
"""
from vistafetch.model.asset import Bond, Fund, Index, PreciousMetal, Stock
from vistafetch.model.search_result import SearchResult

__all__ = [
    "Bond",
    "Fund",
    "Index",
    "PreciousMetal",
    "Stock",
    "SearchResult",
]



"""Configure logging for library."""
import logging

from rich.logging import RichHandler

__all__ = [
    "set_up_logging",
]


def set_up_logging() -> None:
    """Configure logging."""
    log_format = "%(message)s"
    logging.basicConfig(
        level="NOTSET", format=log_format, datefmt="[%X]", handlers=[RichHandler()]
    )

    logging.debug("Logging set up successfully.")



"""Constant variables relevant throughout the entire library.

Attributes
----------
ONVISTA_API_BASE_URL: base URL of the Onvista API
ONVISTA_API_TIMEZONE: datetime timezone that is used by the API

"""

from datetime import timezone

__all__ = ["ONVISTA_API_BASE_URL", "ONVISTA_API_TIMEZONE"]

ONVISTA_API_BASE_URL = "https://api.onvista.de/api/v1/"
ONVISTA_API_TIMEZONE = timezone.utc



"""Models the supported types of financial assets."""

from enum import Enum


class FinancialAssetType(Enum):
    """Possible types of financial asset types.

    Attributes
    ----------
        BOND: A debt security that represents a loan made by
            an investor to a borrower.
        FUND: An investment fund, e.g., mutual fund, ETF, etc.
        INDEX: A basket of securities representing a particular market or
            a segment of it.
        STOCK: Share of a corporation or company.
        UNKNOWN: Unknown type of financial asset.
            Should only be used for abstract modeling.

    """

    BOND = "BOND"
    FUND = "FUND"
    INDEX = "INDEX"
    METAL = "PRECIOUS_METAL"
    STOCK = "STOCK"
    UNKNOWN = None
